@@ -8,14 +8,14 @@
 #' @param data (optional) data.frame, list or environment in which to evaluate
 #'   the variables in \code{formula} and \code{modelweights}.
 #' @param iter number of combinations of starting parameters which will be tried
-#'   . If a single value is provided, then a shotgun/random-search approach will
-#'   be used to sample starting parameters from a uniform distribution within
-#'   the starting parameter bounds. If a vector of the same length as the number
-#'   of parameters is provided, then a gridstart approach will be used to define
-#'   each combination of that number of equally spaced intervals across each of
-#'   the starting parameter bounds respectively. Thus, c(5,5,5) for three fitted
-#'   parameters yields 125 model fits.  Supplying a vector for \code{iter} will
-#'   override \code{convergence_count}.
+#'   . If a single value is provided, then a shotgun/random-search/lhs approach
+#'   will be used to sample starting parameters from a uniform distribution
+#'   within the starting parameter bounds. If a vector of the same length as the
+#'   number of parameters is provided, then a gridstart approach will be used to
+#'   define each combination of that number of equally spaced intervals across
+#'   each of the starting parameter bounds respectively. Thus, c(5,5,5) for
+#'   three fitted parameters yields 125 model fits.  Supplying a vector for
+#'   \code{iter} will override \code{convergence_count}.
 #' @param start_lower lower boundaries for the start parameters. If missing,
 #'   this will default to -1e+10.
 #' @param start_upper upper boundaries for the start parameters. If missing,
@@ -38,6 +38,13 @@
 #' @param modelweights Optional model weights for the nls. If \code{data} is
 #'   specified, then this argument should be the name of the numeric weights
 #'   vector within the \code{data} object.
+#' @param lhstype Method to use for Latin Hypercube Sampling using \code{\link[lhs]{lhs}}. Choice of
+#' \code{"random"} (simple random lhs, fast), \code{"improved"} (lhs with
+#' optimised euclidean distance between points, medium speed), \code{"maximin"}
+#' (lhs with maximised minimum distance between points, medium speed), or
+#' \code{"genetic"} (lhs optimised to the S optimilaity criterion using a
+#' genetic algorithm, slow). If not set, a purely random (shotgun) approach is
+#' taken. Only used if \code{iter} is a single number.
 #' @param \dots Extra arguments to pass to \code{\link[minpack.lm]{nlsLM}} if
 #'   necessary.
 #' @return returns a nls object of the best estimated model fit.
@@ -46,8 +53,9 @@
 #'   represent upper and lower boundaries for parameter estimates.
 #' @author Daniel Padfield
 #' @author Granville Matheson
+#' @author Francis Windram
 #' @seealso \code{\link[minpack.lm]{nlsLM}} for details on additional arguments
-#'   to pass to the nlsLM function.
+#'   to pass to the nlsLM function. \code{\link[lhs]{lhs}} for details of Latin Hypercube Sampling
 #' @examples
 #' # load in data
 #'
@@ -83,7 +91,7 @@ nls_multstart <-
   # arguments needed for nls_multstart ####
   function(formula, data = parent.frame(), iter, start_lower, start_upper,
            supp_errors = c("Y", "N"), convergence_count = 100, control,
-           modelweights, ...) {
+           modelweights, lhstype=c("none", "random", "improved", "maximin", "genetic"), ...) {
 
     # set default values
     if (missing(supp_errors)) {
@@ -99,21 +107,20 @@ nls_multstart <-
       params_est <- all.vars(formula[[3]])[!all.vars(formula[[3]]) %in% names(data)]
       params_dep <- all.vars(formula[[2]])
     } else {
-      stop("data should be a data.frame, list or an environment")
+      cli::cli_abort(c("data should be a {.cls data.frame}, {.cls list} or an {.cls environment}",
+      "x"="You've provided: {.cls {class(data)}}"))
     }
 
     # set up parameter boundaries ####
 
     if (missing(start_lower) || missing(start_upper)) {
-      cat(
-        "No boundaries specified for the starting values of sought parameters. \n",
-        "Default values of +/- 1e+10 will be used. This is likely \n",
-        "to slow the process of finding the best model. \n"
-      )
-      r <- readline("Continue with default values [y/n]? ")
+      cli::cli_warn(c("!"="No boundaries specified for the starting values of sought parameters.",
+                      " "="Default values of +/- {.val {1e+10}} will be used. This is likely to slow the process of finding the best model.",
+                      " "="Continue with default values [Y/n]?"))
+      r <- readline("? ")
 
       if (tolower(r) == "n") {
-        stop("Please enter upper and lower parameter boundaries as start_lower and start_upper in function argument.")
+        cli::cli_abort("Please enter upper and lower parameter boundaries as start_lower and start_upper in the function arguments.")
       }
     }
 
@@ -121,15 +128,18 @@ nls_multstart <-
     if (missing(start_upper)) start_upper <- rep(10 ^ 10, length(params_est))
 
     if (length(start_lower) != length(params_est) || length(start_upper) != length(params_est)) {
-      stop(paste0(
-        "There must be as many parameter starting bounds as there are parameters.\n\n",
-        "Parameters to estimate:              ", paste(params_est, collapse=" "),"\n",
-        "Parameters supplied as input data:   ", paste(params_ind, collapse=" "),"\n",
-        "Number of parameters in start_lower: ", length(start_lower),"\n",
-        "Number of parameters in start_upper: ", length(start_upper),"\n\n",
-        "If there are parameters listed as input data which should be estimated,\n",
-        "this can be caused by variables in the environment/data with the same\n",
-        "name as model parameters."))
+      cli::cli_abort(c(
+        "x"="There must be as many parameter starting bounds as there are parameters.",
+        "",
+        "*"="Parameters to estimate: {.val {params_est}}",
+        "*"="Parameters supplied as input data: {.val {params_ind}}",
+        "*"="Number of parameters in start_lower: {.val {length(start_lower)}}",
+        "*"="Number of parameters in start_upper: {.val {length(start_upper)}}",
+        "",
+        "!"="If there are parameters listed as input data which should be estimated,",
+        " "="this can be caused by variables in the environment/data with the same name as model parameters.",
+        ""
+      ))
     }
 
     params_bds <- data.frame(
@@ -155,10 +165,7 @@ nls_multstart <-
     }
 
     if ("modelweights" %in% all.vars(formula)) {
-      stop(paste0(
-        "The variable name 'modelweights' is reserved for model weights. Please change the name\n",
-        "of this variable"
-      ))
+      cli::cli_abort(c("x"="The variable name {.var modelweights} is reserved for model weights.","!"="Please change the name of this variable."))
     }
 
     if (missing(modelweights)) {
@@ -173,26 +180,38 @@ nls_multstart <-
     } else if (length(iter) == nrow(params_bds)) {
       multistart_type <- "gridstart"
     } else {
-      stop(paste0(
-        "iter should be of length 1 for shotgun approach and of the same length as the\n",
-        "number of parameters for the gridstart approach."
-      ))
+      cli::cli_abort(c("x"="{.arg iter} should be a single number for the shotgun or lhs approaches and of the same length as the number of parameters for the gridstart approach.",
+                       "*"="Number of parameters: {.val {nrow(params_bds)}}"))
     }
 
     if (multistart_type == "gridstart" && convergence_count != FALSE) {
-      warning("A gridstart approach cannot be applied with convergence_count. Convergence count will be set to FALSE")
+      cli::cli_warn(c("!"="A gridstart approach cannot be applied with {.arg convergence_count}.","i"="Convergence count will be set to {.val {FALSE}}"))
       convergence_count <- FALSE
     }
 
+    lhstype <- rlang::arg_match(lhstype)
 
     # set up start values ####
 
     ## SHOTGUN ##
 
     if (multistart_type == "shotgun") {
-      strt <- purrr::map2(params_bds$low.bds, params_bds$high.bds, ~runif(iter, .x, .y))
-      names(strt) <- params_bds$param
-      strt <- dplyr::bind_rows(strt)
+      if (lhstype == "none") {
+        strt <- purrr::map2(params_bds$low.bds, params_bds$high.bds, ~runif(iter, .x, .y))
+        names(strt) <- params_bds$param
+        strt <- dplyr::bind_rows(strt)
+      } else {
+        lhs_mapper <- switch(lhstype,
+                             random=lhs::randomLHS,  # ~400us for 500x5
+                             improved=lhs::improvedLHS, # ~135ms
+                             maximin=lhs::maximinLHS,  # ~135ms
+                             genetic=lhs::geneticLHS,  # ~500ms
+                             {cli::cli_warn(c("!"="Not a known lhs approach, defaulting to {.val {'random'}}"));lhs::randomLHS})
+        lhs_map <- lhs_mapper(n=iter, k=nrow(params_bds))
+        rmat <- apply(params_bds[,2:3], 1, function(x) {diff(range(x))})
+        strt <- tibble::as_tibble(t(apply(lhs_map, 1, function(x) {(x * rmat) + params_bds$low.bds})))
+        colnames(strt) <- params_bds$param
+      }
     }
 
 
@@ -252,7 +271,9 @@ nls_multstart <-
         else {
           count <- ifelse(stored_AIC <= stats::AIC(fit), count + 1, 0)
         }
-        if (count == convergence_count) break
+        if (count == convergence_count) {
+          break
+        }
 
         if (!is.null(fit) && stored_AIC > stats::AIC(fit)) {
           stored_AIC <- stats::AIC(fit)
